from django.forms import ValidationError
from django.forms.fields import CharField
from django.utils.translation import gettext_lazy as _

from .validators import validate_cpf, validate_cnpj
from .utils import is_equal


class InvalidValuesField:
    invalid_values = ()

    def clean(self, value):
        value = super().clean(value)
        if value in self.invalid_values:
            raise ValidationError(_("Invalid number."))
        return value


class CPFField(InvalidValuesField, CharField):
    invalid_values = ('00000000191',)

    def __init__(self, max_length=14, min_length=11, *args, **kwargs):
        super().__init__(max_length=max_length, min_length=min_length, *args, **kwargs)

    def clean(self, value):
        value = super().clean(value)
        code = validate_cpf(value)
        if not code or is_equal(code):
            raise ValidationError(_("Invalid CPF number."))
        return code


class CNPJField(InvalidValuesField, CharField):
    invalid_values = (
        '00000000000000', '22222222000191', '33333333000191', '44444444000191',
        '55555555000191', '66666666000191', '77777777000191', '88888888000191',
        '99999999000191'
    )

    def __init__(self, min_length=14, max_length=18, *args, **kwargs):
        super().__init__(max_length=max_length, min_length=min_length, *args, **kwargs)

    def clean(self, value):
        value = super().clean(value)
        code = validate_cnpj(value)
        if not code or is_equal(code):
            raise ValidationError(_("Invalid CNPJ number."))
        return code




from .generic import SlugModel, TimestampedModel, UUIDModel, SerializerModel  # noqa
from .signals import SignalsModel  # noqa
from .history import HistoryModel  # noqa

__all__ = [
    'SlugModel', 'TimestampedModel', 'UUIDModel', 'SerializerModel',
    'SignalsModel', 'HistoryModel'
]



from django.core.validators import EMPTY_VALUES

from .utils import is_equal
from .utils.string import digits_only

SEQ_VALUES = [
    '00000000000',
    '11111111111',
    '22222222222',
    '33333333333',
    '44444444444',
    '55555555555',
    '66666666666',
    '77777777777',
    '88888888888',
    '99999999999'
]


def dv_maker(v):
    if v >= 2:
        return 11 - v
    return 0


def validate_cpf(value):
    """
    Value can be either a string in the format XXX.XXX.XXX-XX or
    an 11-digit number.
    """
    if value in EMPTY_VALUES or value in SEQ_VALUES:
        return False
    orig_value = value[:]
    if not value.isdigit():
        value = digits_only(value)
        if not value:
            return False

    if is_equal(value):
        return False

    if len(value) != 11:
        return False
    orig_dv = value[-2:]

    r1 = range(10, 1, -1)
    new_1dv = sum([i * int(value[idx]) for idx, i in enumerate(r1)])
    new_1dv = dv_maker(new_1dv % 11)
    value = value[:-2] + str(new_1dv) + value[-1]
    r2 = range(11, 1, -1)
    new_2dv = sum([i * int(value[idx]) for idx, i in enumerate(r2)])
    new_2dv = dv_maker(new_2dv % 11)
    value = value[:-1] + str(new_2dv)
    if value[-2:] != orig_dv:
        return False
    if value.count(value[0]) == 11:
        return False
    return orig_value


def validate_cnpj(value):
    """
    Value can be either a string in the format XX.XXX.XXX/XXXX-XX or
    a group of 14 characters.
    """
    if value in EMPTY_VALUES:
        return ''
    orig_value = value[:]
    if not value.isdigit():
        value = digits_only(value)
        if not value:
            return False

    if is_equal(value):
        return False

    if len(value) != 14:
        return False
    orig_dv = value[-2:]

    list1 = list(range(5, 1, -1))
    list2 = list(range(9, 1, -1))
    new_1dv = sum([i * int(value[idx]) for idx, i in enumerate(list1 + list2)])
    new_1dv = dv_maker(new_1dv % 11)
    value = value[:-2] + str(new_1dv) + value[-1]
    list3 = list(range(6, 1, -1))
    new_2dv = sum([i * int(value[idx]) for idx, i in enumerate(list3 + list2)])
    new_2dv = dv_maker(new_2dv % 11)
    value = value[:-1] + str(new_2dv)
    if value[-2:] != orig_dv:
        return False
    return orig_value



from django.db import models, transaction
from django.db.models import QuerySet
from django.db.models.deletion import Collector
from django.utils import timezone

REPR_OUTPUT_SIZE = 20


class SignalsManager(models.Manager):

    def create(self, **kwargs):
        model_instance = self.initialize_model_instance(**kwargs)
        with transaction.atomic():
            model_instance.save()
        return model_instance

    def initialize_model_instance(self, **kwargs):
        return self.model(**kwargs)


class SoftDeleteQuerySet(QuerySet):
    def delete(self):
        return super().update(deleted_at=timezone.now(), is_deleted=True)

    def hard_delete(self):
        """Delete the records in the current QuerySet."""
        self._not_support_combined_queries('delete')
        assert not self.query.is_sliced, \
            "Cannot use 'limit' or 'offset' with delete."

        if self._fields is not None:
            raise TypeError("Cannot call delete() after .values() or .values_list()")

        del_query = self._chain()

        # The delete is actually 2 queries - one to find related objects,
        # and one to delete. Make sure that the discovery of related
        # objects is performed on the same database as the deletion.
        del_query._for_write = True

        # Disable non-supported fields.
        del_query.query.select_for_update = False
        del_query.query.select_related = False
        del_query.query.clear_ordering(force_empty=True)

        collector = Collector(using=del_query.db)
        collector.collect(del_query)
        deleted, _rows_count = collector.delete()

        # Clear the result cache, in case this QuerySet gets reused.
        self._result_cache = None
        return deleted, _rows_count

    def restore(self):
        return super().update(deleted_at=None, is_deleted=False)

    def __repr__(self):
        data = list(self[:REPR_OUTPUT_SIZE + 1])
        if len(data) > REPR_OUTPUT_SIZE:
            data[-1] = "...(remaining elements truncated)..."
        return '<QuerySet %r>' % data


class SoftDeleteSignalsManager(SignalsManager):
    def __init__(self, *args, **kwargs):
        self.show_deleted = kwargs.pop('show_deleted', False)
        super().__init__(*args, **kwargs)

    def get_queryset(self):
        if self.show_deleted:
            return SoftDeleteQuerySet(self.model, using=self._db)
        return SoftDeleteQuerySet(self.model, using=self._db).exclude(is_deleted=True)

    def delete(self):
        return self.get_queryset().delete()

    def hard_delete(self):
        return self.get_queryset().hard_delete()

    def restore(self):
        return self.get_queryset().restore()

    def filter(self, *args, **kwargs):
        if 'is_deleted' in kwargs:
            qs = SoftDeleteQuerySet(self.model, using=self._db)
            return qs.filter(*args, **kwargs)
        return super().filter(*args, **kwargs)

    def trash(self):
        return self.filter(is_deleted=True)



import re
import unicodedata


def digits_only(raw_number):
    return re.sub(r"[^0-9]", "", raw_number)


def is_equal(words):
    return all(c == words[i - 1] for i, c in enumerate(words) if i > 0)


def remove_special_characters(text):
    cleaned_text = re.sub(r'([^\s\w]|_)+', '', text.strip())
    return remove_accents(cleaned_text)


def remove_accents(text):
    return ''.join(c for c in unicodedata.normalize('NFD', text)
                   if unicodedata.category(c) != 'Mn')



from .generic import TimestampedModel
from .signals import SignalsModel
from .exceptions import HistoryModelNotSetError


class HistoryModel(SignalsModel, TimestampedModel):
    history_model = None
    history_parent_field_name = 'parent'

    def __init__(self, *args, **kwargs):
        if not self.history_model:
            raise HistoryModelNotSetError(
                "You should set the history_model attribute of {}".format(type(self).__name__)
            )

        super().__init__(*args, **kwargs)

    def save_history(self):
        data = {self.history_parent_field_name: self}

        history_model_fields = tuple(field.name for field in self.history_model._meta.fields)

        for field in self._meta.fields:
            if field.name in ('id', 'created_at', 'updated_at'):
                continue

            if field.name in history_model_fields:
                value = getattr(self, field.name)
                data[field.name] = value

        self.history_model.objects.create(**data)

    def post_save_history(self, context):
        self.save_history()



import uuid

from django.core.validators import RegexValidator
from django.db import models


class UUIDPrimaryKeyField(models.UUIDField):

    def __init__(self, *args, **kwargs):
        kwargs['primary_key'] = True
        kwargs['unique'] = True
        kwargs['editable'] = False
        super().__init__(*args, **kwargs)

    def pre_save(self, model_instance, add):
        value = super().pre_save(model_instance, add)

        if value is None:
            value = uuid.uuid4()
            setattr(model_instance, self.attname, value)

        return value


class CharFieldDigitsOnly(models.CharField):
    default_validators = [RegexValidator(r'^([\s\d]+)$', 'Only digits characters')]



import hashlib
import random
import string
import time
from datetime import datetime, timedelta

from django.utils.crypto import get_random_string


def generate_random_code(length=10):
    allowed = string.ascii_uppercase + string.digits
    code = get_random_string(length=length, allowed_chars=allowed)
    return code


def generate_md5_hashcode(key_word):
    keyword = '{}-{}'.format(key_word, time.time())
    hashcode = hashlib.md5(keyword.encode('utf-8')).hexdigest()
    return hashcode


def generate_datetime(min_year=1900, max_year=datetime.now().year):
    """Generate a datetime."""
    start = datetime(min_year, 1, 1, 00, 00, 00)
    years = max_year - min_year + 1
    end = start + timedelta(days=365 * years)
    return start + (end - start) * random.random()


def generate_cpf():
    cpf = [random.randint(0, 9) for x in range(9)]

    for _ in range(2):
        val = sum([(len(cpf) + 1 - i) * v for i, v in enumerate(cpf)]) % 11

        cpf.append(11 - val if val > 1 else 0)

    return '%s%s%s.%s%s%s.%s%s%s-%s%s' % tuple(cpf)


def generate_cnpj():
    def calculate_special_digit(l):
        digit = 0

        for i, v in enumerate(l):
            digit += v * (i % 8 + 2)

        digit = 11 - digit % 11

        return digit if digit < 10 else 0

    cnpj = [1, 0, 0, 0] + [random.randint(0, 9) for x in range(8)]

    for _ in range(2):
        cnpj = [calculate_special_digit(cnpj)] + cnpj

    return '%s%s.%s%s%s.%s%s%s/%s%s%s%s-%s%s' % tuple(cnpj[::-1])



#!/usr/bin/env python
# encoding: utf-8

from . import generators  # noqa
from .generic import find_path, get_version_from_changes  # noqa
from .string import remove_special_characters, remove_accents, is_equal, digits_only  # noqa

__all__ = ['remove_special_characters', 'remove_accents', 'is_equal', 'digits_only',
           'find_path', 'get_version_from_changes']



import inspect

from django.db import models, transaction
from django.utils import timezone

from .generic import SerializerModel
from .managers import SignalsManager, SoftDeleteSignalsManager


class SignalsModel(SerializerModel):
    SOFT_DELETE = False

    class Meta:
        abstract = True

    objects = SignalsManager()

    def get_context(self, **kwargs):
        force_insert = kwargs.get('force_insert', False)
        creation_conditions = (
            self.id is None,
            force_insert is True
        )
        context = {'is_creation': any(creation_conditions)}
        context.update(kwargs)
        return context

    def trigger_event(self, event_name, context):
        for attribute in dir(self):
            if attribute.startswith(event_name):
                method = getattr(self, attribute)
                if inspect.ismethod(method):
                    method(context)

    def save(self, *args, **kwargs):
        force_insert = kwargs.get('force_insert', False)
        context = self.get_context(force_insert=force_insert)

        with transaction.atomic():
            self.trigger_event('pre_save', context)
            super().save(*args, **kwargs)
            self.trigger_event('post_save', context)

    def delete(self, *args, **kwargs):
        context = self.get_context()

        if self.SOFT_DELETE and not kwargs.pop('hard_delete', False):
            self.deleted_at = timezone.now()
            self.is_deleted = True
            self.save()
            return

        with transaction.atomic():
            self.trigger_event('pre_delete', context)
            super().delete(*args, **kwargs)
            self.trigger_event('post_delete', context)


class SoftDeleteSignalModel(SignalsModel):
    SOFT_DELETE = True
    deleted_at = models.DateTimeField(blank=True, null=True)
    is_deleted = models.BooleanField(default=False)

    objects = SoftDeleteSignalsManager()
    all_objects = SoftDeleteSignalsManager(show_deleted=True)

    class Meta:
        abstract = True

    def hard_delete(self):
        super().delete(hard_delete=True)

    def restore(self):
        self.deleted_at = None
        self.is_deleted = False
        self.save()



import codecs
import fnmatch
import os
import re
from unipath import Path


def find_path(pattern, path, last_folder_only=False, ignored_dirs=''):
    ignored_dirs = ignored_dirs.split(',')
    result = []

    for root, dirs, files in os.walk(path):
        dirs[:] = [d for d in dirs if d not in ignored_dirs]

        for name in files:
            if fnmatch.fnmatch(name, pattern):
                result.append(os.path.join(root, name))

    if not result:
        return ''

    final_path = Path(result[0]).ancestor(1)
    relative_path = final_path.replace(str(path), '')

    if last_folder_only:
        return relative_path.split('/')[-1]
    return relative_path[1:] if relative_path.startswith('/') else relative_path


def get_version_from_changes(project_root=''):
    default_version = '0.0.0'
    current_version = ''
    changes = os.path.join(project_root, "CHANGES.rst")
    pattern = r'^(?P<version>[0-9]+.[0-9]+(.[0-9]+)?)'
    if not os.path.exists(changes):
        return default_version

    with codecs.open(changes, encoding='utf-8') as changes:
        for line in changes:
            match = re.match(pattern, line)
            if match:
                current_version = match.group("version")
                break
    return current_version or default_version



from functools import partial
from uuid import UUID

from django.db import models
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from rest_framework.serializers import ModelSerializer

from django_models.fields import UUIDPrimaryKeyField
from django_models.utils.generators import generate_random_code

generate_code = partial(generate_random_code, length=16)


class BaseModel(models.Model):
    class Meta:
        abstract = True


class ActiveModel(models.Model):
    active = models.BooleanField(_('Active'), default=True)

    class Meta:
        abstract = True


class CodeModel(models.Model):
    code = models.CharField(
        max_length=32,
        default=generate_code,
        verbose_name=_('Model code'),
        unique=True,
    )

    class Meta:
        abstract = True


class SerializerModel(BaseModel):

    @cached_property
    def serializer(self):
        class SelfSerializer(ModelSerializer):
            class Meta:
                pass

        SelfSerializer.Meta.model = self
        SelfSerializer.Meta.fields = '__all__'
        return SelfSerializer

    def serialize(self):
        data = self.serializer(self).data

        for key, value in data.items():
            if isinstance(value, UUID):
                data[key] = str(value)
        return data

    class Meta:
        abstract = True


class SlugModel(BaseModel):
    slug = models.SlugField(max_length=16)

    class Meta:
        abstract = True


class TimestampedModel(BaseModel):
    created_at = models.DateTimeField(
        auto_now_add=True, verbose_name=_('Created at')
    )
    updated_at = models.DateTimeField(
        auto_now=True, verbose_name=_('Updated at')
    )

    class Meta:
        abstract = True


class UUIDModel(BaseModel):
    id = UUIDPrimaryKeyField()

    class Meta:
        abstract = True
